;;;; scm2mach.scm: Program for translating SIMSYNCH code to MACHXL -*-scheme-*-
;;; Copyright (C) 1991, 1992, 1993, 1995, 1996, 1997 Aubrey Jaffer.

;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;; $Header: /usr/local/cvsroot/synch/scm2mach.scm,v 1.12 1998/06/02 15:03:18 radey Exp $

;;; REPORT an error or warning
(define report
  (lambda args
    (display *pds-output-name*)
    (display ": In function `")
    (display *current-signal-name*)
    (display "': ")
    (display *pds-output-line-number*)
    (display ": warning: ")
    (apply qreport args)))

(define qreport
  (lambda args
    (for-each (lambda (x) (write x) (display #\ )) args)
    (newline)))

(require 'common-list-functions)
(require 'fluid-let)
(require 'printf)
;;(require 'debug) (set! *qp-width* 100) (define qreport qpn)

;;; This allows us to test without generating files
(define *pds-output* (current-output-port))
(define *pds-output-name* "?")

;;; *index* is the bit of an integer -- not a subscript.
(define *index* #f)			;dynamic variable for looping vects.
(define *range* #f)			;dynamic variable for numeric constants.
(define *current-signal-name* #f)
(define *pds-output-line-number* 0)
(define CONTLINE -80)

(define doing-constant? #f)

;;; OUT indents and displays the arguments
(define (out indent . args)
  (cond ((>= indent 0)
	 (newline *pds-output*)
	 (set! *pds-output-line-number* (+ 1 *pds-output-line-number*))
	 (do ((j indent (- j 8)))
	     ((> 8 j)
	      (do ((i j (- i 1)))
		  ((>= 0 i))
		(display #\  *pds-output*)))
	   (display #\	 *pds-output*))))
  (for-each (lambda (a)
	      (cond ((eq? '? a) (display "?" *pds-output*))
		    ((symbol? a) (translate-name a *pds-output*))
		    (else (display a *pds-output*))))
	    args))

;;; TRANSLATE-NAME removes or translates characters from name and prints to port
(define (translate-name name port)
  (define visible? #f)
  (for-each
   (lambda (c)
     (let ((tc (cond ((char-alphabetic? c) c)
		     ((char-numeric? c) c)
		     ((char=? c #\.) #\.)
		     ((char=? c #\[) #\[)
		     ((char=? c #\]) #\])
		     ((char=? c #\-) #\_)
		     ((char=? c #\_) #\_)
		     ((char=? c #\?) "_P")
		     (else #f))))
       (cond (tc (set! visible? #t)
		 (display tc port)))))
   (string->list (symbol->string name)))
  (if (not visible?) (report "C-invisible symbol?" name)))

(define (outcomment indent str)
  (out indent ";" str)
  (out indent))

(define (out-pins pin-nums . rest)
  (cond ((eq? '? pin-nums)
	 (apply out 0 "PIN " '? " " rest))
	((number? pin-nums)
	 (apply out 0 "PIN " (if (zero? pin-nums) "?" pin-nums) " " rest))
	((and (list? pin-nums)
	      (every (lambda (obj) (and (number? obj) (not (zero? obj))))
		     pin-nums))
	 (out 0 "PIN " (car pin-nums))
	 (for-each (lambda (num) (out CONTLINE "," num)) (cdr pin-nums))
	 (apply out CONTLINE " " rest))
	((and (list? pin-nums) (pair? rest) (vector-bounds (car rest)))
	 =>
	 (lambda (bounds)
	   (let* ((first (cadr bounds))
		  (last (caddr bounds))
		  (range (- last first))
		  (incer (sign range)))
	     (do ((i 0 (+ incer i)))
		 ((positive? (* incer (- i range))))
	       (apply out-pins
		      (list-ref pin-nums (abs i))
		      (refvectorize (car rest) (+ i first)) (cdr rest))))))
	(else
	 (apply report "funny pins?" pin-nums rest))))

(define (mach:revectorize sym ind1 ind2)
  (revectorize ".." sym ind1 ind2))

(define (descmfilify file)
  (let ((sl (string-length file)))
  (cond ((< sl 4) file)
	((string-ci=? (substring file (- sl 4) sl) ".scm")
	 (substring file 0 (- sl 4)))
	(else file))))

;;; TRANSLATE files.
(define (translate-design-tag-from-files design tag files)
  (for-each (lambda (f) (translate1 f design tag ".pds")) files))

;;; TRANSLATE1 - translate file.scm to file.suffix
(define (translate1 file design tag suffix)
  (define ofile
    (string-append (symbol->string design) "-" (symbol->string tag) suffix))
  (printf "translating %s -> %s for ptag %s\n" file ofile tag)
  (set! *pds-output* (open-file ofile "w"))
  (set! *pds-output-name* ofile)
  (set! *pds-output-line-number* 0)
  (translate-tag design tag)
  (close-output-port *pds-output*)
  (set! *pds-output* (current-output-port)))

(define (numbers-in-expression? tree)
  (let walk ((tree tree))
    (cond ((number? tree))
	  ((not (pair? tree)) #f)
	  ((not (list? tree)) #f)
	  ((memq (car tree) '(LOGBIT? = BIT-EXTRACT ASH)) #f)
	  ((some number? (cdr tree)))
	  (else (some walk (cdr tree))))))

(define (registered-input? name)
  (and (symbol? name)
       (let ((name (symbol->string name)))
	 (and (> (string-length name) 4)
	      (string=? "pre:" (substring name 0 4))
	      (string->symbol (substring name 4 (string-length name)))))))

(define (generate-pds-header design tag)
  (require 'posix-time)
  (let* ((ptag (or (ptag->pld tag) "PALCE630"))
	 (signature (ptag->sig tag)))
    (newline *pds-output*)
    (fprintf *pds-output* "TITLE     %s\n" (symbol-append design '- tag))
    (fprintf *pds-output* "PATTERN   %s\n" (or time-stamp:configuration ""))
    (fprintf *pds-output* "REVISION  %a\n" time-stamp:revision)
    (and time-stamp:author
	 (fprintf *pds-output* "AUTHOR    %s\n" time-stamp:author))
    (and time-stamp:company
	 (fprintf *pds-output* "COMPANY   %s\n" time-stamp:company))
    (fprintf *pds-output* "DATE      %s\n" (ctime time-stamp:time))
    (fprintf *pds-output* "CHIP  _%s  %s\n" tag ptag)
    (cond ((and signature
		(> (string-length ptag) 4)
		(string-ci=? (substring ptag 0 4) "mach"))
	   (if (not (eqv? signature (logand #xffffffff signature)))
	       (report "signature value out of range"
		       (number->string signature 16))
	       (fprintf *pds-output* "SIGNATURE #h%x\n" signature))))))

;;; TRANSLATE-TAG - translate top level forms.
(define (translate-tag design tag)
  (generate-pds-header design tag)
  ((pin-table 'for-each-row)
   (lambda (pin-row)
     (apply
      (lambda (ptag name pin-type pin-number output-enable paired-input)
	(case pin-type
	  ((output input/output wire-and wire-or internal)
	   (cond ((not (or (not ptag) (eq? ptag tag))))
		 ((not (boolean? paired-input))
		  (cond ((equal? name paired-input)
			 (out-pins pin-number name " REGISTERED"))
			(else
			 (out 0 "NODE ? ")
			 (translate-exp CONTLINE name)
			 (out CONTLINE " REGISTERED PAIR ")
			 (translate-exp CONTLINE paired-input)
			 (out-pins pin-number)
			 (translate-exp CONTLINE paired-input)
			 (out CONTLINE " REGISTERED"))))
		 (paired-input
		  (out-pins pin-number)
		  (translate-exp CONTLINE name)
		  (out CONTLINE " COMBINATORIAL"))
		 (else
		  (out 0 "NODE ? ")
		  (translate-exp CONTLINE name)
		  (out CONTLINE " REGISTERED"))))))
      pin-row)))

  (display "\n\n; Explicit Input-only Pins\n" *pds-output*)
  ((pin-table 'for-each-row)
   (lambda (pin-row)
     (apply
      (lambda (ptag name pin-type pin-number output-enable paired-input)
	(case pin-type
	  ((input)
	   (cond ((eq? ptag tag)
		  (let ((aftername (registered-input? name)))
		    (if aftername
			(out-pins pin-number name " PAIR " aftername)
			(out-pins pin-number name " COMBINATORIAL"))))))))
      pin-row)))

  (display "\n\n; STRING definitions for MACHXL\n" *pds-output*)
  (fluid-let ((doing-constant? #t))
    ((constant-table 'for-each-row)
     (lambda (constant-row)
       (apply
	(lambda (name value)
	  (cond ((or (boolean? value) (number? value))
		 (translate-defmacro-top name value))))
	constant-row))))
  ((signal-table 'for-each-row)
   (lambda (signal-row)
     (apply
      (lambda (block name defining-ptag signal-type initial-state next-function order)
	(case signal-type
	  ((macro)
	   (and (or (not defining-ptag) (eq? tag defining-ptag))
		(translate-vectorized-defmacro name next-function)))))
      signal-row)))

  (display "\n\nEQUATIONS\n" *pds-output*)
  ((signal-table 'for-each-row)
   (lambda (signal-row)
     (apply
      (lambda (block name defining-ptag signal-type initial-state next-function order)
	(case signal-type
	  ((registered registered-input combinatorial)
	   (and
	    (or (not defining-ptag) (eq? tag defining-ptag))
	    (pin-table:row-retrieve tag name)
	    (apply (lambda (pin-type pin-number output-enable paired-input)
		     (cond
		      ((not (eq? paired-input #t))
		       (out 0 name ".CLKF = " block "_CLOCK")
		       (set! initial-state (ceval initial-state))
		       (cond (;;(equal? name initial-state)
			      (not (or (number? initial-state)
				       (boolean? initial-state)))
			      (cond (output-enable
				     (display "self-initing ")
				     (display name) (newline)
				     (out 0 name ".RSTF = GND")
				     (out 0 name ".SETF = GND"))
				    (else
				     (display "Input-register ")
				     (display name) (newline))))
			     ((and (number? initial-state) (zero? initial-state))
			      (translate-re/set block name #f))
			     ((number? initial-state)
			      (let* ((bounds (vector-bounds name))
				     (first (cadr bounds))
				     (last (caddr bounds))
				     (range (- last first)))
				(cond ((= initial-state
					  (+ -1 (ash 2 (abs range))))
				       (translate-re/set block name #t))
				      (else
				       (for-each-index
					(lambda (idx)
					  (translate-re/set
					   tag
					   (refvectorize name idx)
					   (logbit? *index* initial-state)))
					bounds)))))
			     ((boolean? initial-state)
			      (translate-re/set block name initial-state))
			     (else
			      (report "funny initial value for" name ':
				      initial-state)
			      (translate-re/set block name #t)))))
		     (cond (paired-input
			    (out 0)
			    (translate-exp
			     0 (if (eq? paired-input #t) name paired-input))
			    (out CONTLINE ".TRST = ")
			    (translate-exp 8 output-enable)))
		     ;;(translate-vectorized-assignment (list 'define name next-function))
		     (fluid-let ((*block* block))
		       (translate-top
			(list 'define name next-function))))
		   (cddr (pin-table:row-retrieve tag name)))))))
      signal-row))))

(define (translate-vectorized-defmacro name next-function)
  (cond ((and (numbers-in-expression? next-function) (vector-bounds name))
	 => (lambda (bounds)
	      (report 'machxl 'cannot 'translate-vectorized-defmacro stepper)
	      (for-each-index (lambda (idx)
				(translate-defmacro-top (refvectorize name idx)
							next-function))
			      bounds)))
	(else (translate-defmacro-top name next-function))))

(define (translate-defmacro-top var def)
  (set! *current-signal-name* var)
  (cond ((number? def)
	 (out 0 "STRING " var " '")	;name
	 (translate-exp 8 def)
	 (out CONTLINE "'"))
	(else
	 (out 0 "STRING " var " '(")	;name
	 (translate-exp 8 def)
	 (out CONTLINE ")'")))
  (out 0))

(define (translate-re/set block name bool)
  (cond (bool				;(xor bool (negated? name))
	 (out 0 name ".RSTF = GND")
	 (out 0 name ".SETF = /" block "_RESET_"))
	(else
	 (out 0 name ".RSTF = /" block "_RESET_")
	 (out 0 name ".SETF = GND"))))

(define (for-each-index proc bounds)
  (let* ((first (cadr bounds))
	 (last (caddr bounds))
	 (range (- last first)))
    (if (positive? range)
	(do ((i 0 (+ 1 i))
	     (j first (+ 1 j)))
	    ((> i range))
	  (fluid-let ((*index* i))
	    ;;(print '+range i j)
	    (proc j)))
	(do ((i (- range) (+ -1 i))
	     (j first (+ -1 j)))
	    ((negative? i))
	  (fluid-let ((*index* i))
	    ;;(print '-range i j)
	    (proc j))))))

(define (for-each-value proc bounds)
  (let* ((first (cadr bounds))
	 (last (caddr bounds))
	 (range (- last first)))
    (if (not (zero? (min first last)))
	(report "mapping over non-consecutive values" bounds))
    (set! range (ash 2 (abs range)))
    (do ((i 0 (+ 1 i)))
	((>= i range))
      (fluid-let ((*index* i))
	(proc i)))))

;;; TRANSLATE-TOP - translate top level form sexp.
(define (translate-top sexp)
  (cond ((symbol? sexp) (set! *current-signal-name* sexp))
	((and (pair? sexp) (eq? (car sexp) 'QUOTE))
	 (set! *current-signal-name* (cadr sexp)))
	((string? sexp) (outcomment 0 sexp))
	((not (pair? sexp))
	 (report "top level atom?" sexp))
	(else (case (car sexp)
		((define) (cond ((ifable? (caddr sexp))
				 (fix-polarity (cadr sexp))))
			  (translate-assignment (cadr sexp) 0 (caddr sexp)))
		(else (report "statement not in procedure" sexp)))
	      (set! *current-signal-name* #f)))
  (out 0))

(define (fix-polarity lhs)
  (let ((bounds (vector-bounds lhs)))
    (cond (bounds
	   (out 0 "IF (GND) THEN BEGIN")
	   (out 3 lhs " = #b"
		(make-string (+ 1 (abs (- (caddr bounds) (cadr bounds))))
			     #\1))
	   (out 0 "END"))
	  #+invert-signal-sense
	  ((negated? lhs)
	   (report "inverted polarity doesn't work" lhs)
	   (out 0 "IF (GND) THEN BEGIN")
	   (out 3 lhs " = GND")
	   (out 0 "END"))
	  (else
	   (out 0 "IF (GND) THEN BEGIN")
	   (out 0 lhs " = VCC")
	   (out 0 "END")))))

(define (ifable? rhs)
  (and (pair? rhs)
       (case (car rhs)
	 ((CASE CASEV IF AND NUMBER-OR OR) #t) ; NOT
	 ((SYNCH:PRE) (ifable? (signal-table:get-next-fun *block* (cadr rhs))))
	 ((LOGBIT?) (or (not (number? (cadr rhs)))
			(ifable? (caddr rhs))))
	 (else #f))))

(define (invertify sexp)
  (if (and (pair? sexp) (eq? 'NOT (car sexp)))
      (cadr sexp)
      (list 'NOT sexp)))

(define (translate-assignment lhs indent rhs)
  (set! *current-signal-name* lhs)
  (cond ((ifable? rhs)
	 (case (car rhs)
	   ((LOGBIT?)
	    (cond
	     ((not (number? (cadr rhs)))
	      (let* ((bounds (vector-bounds (cadr rhs)))
		     (clauses '()))
		(for-each-value
		 (lambda (idx)
		   (set!
		    clauses
		    (cons `((,idx)
			    ,(cond ((number? (caddr rhs))
				    (logbit? idx (caddr rhs)))
				   (else `(logbit? ,idx ,(caddr rhs)))))
			  clauses)))
		 bounds)
		;;(pretty-print `(casev ,(cadr rhs) ,@clauses))
		(translate-assignment
		 lhs (+ indent 4) `(casev ,(cadr rhs) ,@clauses))))
	     ((and (pair? (caddr rhs)) (eq? 'IF (caaddr rhs)))
	      (let ((ifexp (caddr rhs)))
		(translate-assignment
		 lhs
		 (+ 2 indent)
		 `(IF ,(cadr ifexp)
		      ,@(map (lambda (se) `(logbit? ,(cadr rhs) ,se))
			     (cddr ifexp))))))
	     ((and (pair? (caddr rhs))
		   (assq (caaddr rhs) '((LOGIOR OR) (LOGAND AND))))
	      => (lambda (pr)
		   (translate-assignment
		    lhs
		    (+ 2 indent)
		    (cons (cadr pr)
			  (map (lambda (se) `(logbit? ,(cadr rhs) ,se))
			       (cdaddr rhs))))))
	     (else (report "logbit assignment not handled" rhs))))
	   ((SYNCH:PRE)
	    (let ((expn (signal-table:get-next-fun *block* (cadr rhs))))
	      (if (not expn) (report "returned #f" rhs))
	      (translate-exp indent expn)))
	   ((CASE CASEV)
	    (out indent "CASE (")
	    (translate-exp (+ 6 indent) (cadr rhs))
	    (out CONTLINE ") BEGIN")
	    (for-each (lambda (clause)
			(translate-case-clause lhs (+ 2 indent) clause))
		      (cddr rhs))
	    (out indent "END")
	    ;;(out indent)
	    #t)
	   ((IF) (out indent "IF (")
		 (translate-exp (+ 2 indent) (cadr rhs))
		 (out CONTLINE ") THEN BEGIN")
		 (translate-assignment lhs (+ 2 indent) (caddr rhs))
		 (out indent "END")
		 (cond ((not (null? (cdddr rhs)))
			(out indent "ELSE BEGIN")
			(translate-assignment lhs (+ 2 indent) (cadddr rhs))
			(out indent "END"))
		       (else (report 'dangling-if? lhs rhs)))
		 #t)
	   ((AND) (translate-assignment
		   lhs
		   indent
		   (case (length rhs)
		     ((1) #t)
		     ((2) (cadr rhs))
		     (else `(if ,(butlast rhs 1) ,(car (last rhs 1)) #f)))))
	   ((OR) (for-each (lambda (c) (translate-assignment lhs indent c))
			   (cdr rhs)))
	   ((NUMBER-OR)
	    ;;(trace-all "scm2mach.scm") (set! *qp-width* 333) (untrace out out-pins translate-name lookup registered-input? numbers-in-expression?) ; translate-exp infix-translate-exp
	    (for-each (lambda (c) (translate-assignment lhs indent c))
		      (cddr rhs)))
	   (else (report "ifable expression not translated" rhs))))
	((and (not *range*)
	      (not *index*)
	      (vector-bounds lhs))
	 =>
	 (lambda (bounds)
	   ;;(print 'translate-assignment 'rhs rhs)
	   (let* ((first (cadr bounds))
		  (last (caddr bounds))
		  (range (- last first))
		  (incer (sign range)))
	     (cond ((or (number? rhs)
			(symbol? rhs)
			(and (pair? rhs)
			     (memq (car rhs) '(BIT-EXTRACT BIT-FIELD))))
		    (fluid-let ((*range* (* incer (+ 1 (abs range)))))
		      (out indent lhs " = ") ;name
		      (translate-exp (+ 8 indent) rhs)
		      ;;(out indent)
		      ))
		   (else
		    (for-each-index
		     (lambda (idx)
		       (translate-assignment (refvectorize lhs idx) indent rhs))
		     bounds))))))
	(else
	 (out indent lhs " = ")		;name
	 (translate-exp (+ 8 indent) rhs)
	 ;;(out indent)
	 )))

(define (translate-case-clause lhs indent clause)
  (cond ((eq? 'else (car clause))
	 (let ((lc (car (last-pair clause))))
	   (cond ((and (pair? lc)
		       (memq (car lc) '(synch:chat synch:warn synch:error))))
		 (else
		  (if (not (member lc '(0 #f)))
		      (report "case else not (0) #f?" lc))
		  (out indent "OTHERWISE: BEGIN")
		  (translate-assignment lhs (+ 10 indent) lc)
		  (out indent "END")))))
	((= 1 (length (car clause)))
	 (out indent (case-eval (caar clause)) ": BEGIN")
	 (translate-assignment lhs (+ 2 indent) (car (last-pair clause)))
	 (out indent "END"))
	(else
	 (for-each (lambda (match)
		     (translate-case-clause
		      lhs indent (cons (list match) (cdr clause))))
		   (car clause)))))

(define (case-eval dat)
  (if (and (pair? dat) (memq (car dat) '(UNQUOTE-SPLICING UNQUOTE)))
      (cadr dat)
      ;;(let ((row (constant-table:row-retrieve (cadr dat)))) (if row (cadr row) dat))
      dat))

;;; LOOKUP - translate from table or return arg as string
(define (lookup arg tab)
  (let* ((p (assq arg tab))
	 (l (if p (cdr p) arg)))
    (if (symbol? l) (symbol->string l) l)))

;;; TRANSLATE-EXP - translate expression
(define (translate-exp indent sexp)
  (cond ((not (pair? sexp))		;atoms
	 (cond ((string? sexp)
		(out CONTLINE #\" sexp #\"))
	       ((boolean? sexp)
		(out CONTLINE (case sexp ((#f) "GND") ((#t) "VCC"))))
	       ((and *index* (symbol? sexp) (vector-bounds sexp))
		(translate-exp indent (if (negative? *index*)
					#f
					`(logbit? ,*index* ,sexp))))
	       ((and *index* (integer? sexp))
		(out CONTLINE (if (and (not (negative? *index*))
				       (logbit? *index* sexp))
				  "VCC" "GND")))
	       ((symbol? sexp) (out CONTLINE sexp))
	       ((and *range* (number? sexp))
		(let* ((num (if (positive? *range*)
				(bit-reverse *range* sexp)
				sexp))
		       (str (number->string num 2)))
		  (out CONTLINE "#B"
		       (make-string (- (abs *range*) (string-length str)) #\0)
		       str)))
	       (doing-constant? (out CONTLINE sexp))
	       (else (report "unexpected datatype:" sexp)
		     (out CONTLINE sexp))))
	(else
	 (case (car sexp)
	   ;; These vectorized functions are for multiple bit signals:
	   ((=)
	    (cond ((not (= 2 (length (cdr sexp))))
		   (report "= should take two arguments" sexp)))
	    (translate-= indent (cadr sexp) (caddr sexp)))
	   ((ZERO?)
	    (translate-= indent 0 (cadr sexp)))
	   ((LOGNOT)
	    (cond ((symbol? (cadr sexp))
		   (out CONTLINE "/")
		   (translate-exp (+ 2 indent) (cadr sexp)))
		  (else
		   (out CONTLINE "/" "(")
		   (translate-exp (+ 2 indent) (cadr sexp))
		   (out CONTLINE ")"))))
       	   ((LOGIOR LOGAND LOGXOR)
	    (infix-translate-exp
	     (lookup (car sexp)
		     '((LOGIOR . " + ")
		       (LOGAND . " * ") (LOGXOR . " :+: ")))
	     indent
	     (cdr sexp)))
	   ((NUMBER-OR) (infix-translate-exp " + " indent (cddr sexp)))
	   ((ASH)
	    (cond (*index* (fluid-let ((*index* (- *index* (caddr sexp))))
			     (translate-exp indent (cadr sexp))))
		  ((number? (caddr sexp)) (apply ash (cdr sexp)))
		  (else (report "ash not handled" sexp))))
	   ((BIT-EXTRACT BIT-FIELD)
	    (let* ((bounds (vector-bounds (cadr sexp)))
		   (lowbnd (if bounds (min (cadr bounds) (caddr bounds)) 0)))
	      (cond ((not bounds)
		     (report "bit extract of non-symbol?" sexp)))
	      (translate-exp
	       CONTLINE
	       (mach:revectorize (cadr sexp)
				 (+ lowbnd (caddr sexp))
				 (+ lowbnd (cadddr sexp) -1)))))
	   ((BITWISE-IF)
	    (let ((mask (cadr sexp))
		  (n0 (caddr sexp))
		  (n1 (cadddr sexp)))
	      (translate-exp
	       CONTLINE
	       `(logand (logand ,mask ,n0)
			(logand (lognot ,mask) ,n1)))))
	   ((LOGBIT?)
	    (cond
	     ((symbol? (cadr sexp))
	      (let* ((bounds (vector-bounds (cadr sexp)))
		     (clauses '()))
		(for-each-value
		 (lambda (idx)
		   (set!
		    clauses
		    (cons `((,idx)
			    ,(cond ((number? (caddr sexp))
				    (logbit? idx (caddr sexp)))
				   (else `(logbit? ,idx ,(caddr sexp)))))
			  clauses)))
		 bounds)
		(pretty-print `(casev ,(cadr sexp) ,@clauses))
		(translate-exp (+ indent 4) `(casev ,(cadr sexp) ,@clauses))))
	     ((and (number? (cadr sexp)) (negative? (cadr sexp))
		   (translate-exp indent #f)))
	     ((number? (caddr sexp))
	      (translate-exp (+ indent 4) (apply logbit? (cdr sexp))))
	     ((symbol? (caddr sexp))
	      (let* ((bounds (vector-bounds (caddr sexp)))
		     (lowbnd (if bounds (min (cadr bounds) (caddr bounds)) 0))
		     (dvs (devectorize (caddr sexp) (+ lowbnd (cadr sexp)))))
		;;(if (not (zero? lowbnd)) (print 'logbit? sexp bounds lowbnd))
		(translate-exp CONTLINE dvs)
		(cond (dvs (out CONTLINE #\[)
			   ;;(translate-exp CONTLINE (+ lowbnd (cadr sexp)))
			   (out CONTLINE (+ lowbnd (cadr sexp)))
			   (out CONTLINE #\])))))
	     ((and (pair? (caddr sexp)) (eq? 'ASH (caaddr sexp)))
	      (let ((ashexp (caddr sexp)))
		(translate-exp CONTLINE
			       `(logbit? ,(- (cadr sexp) (caddr ashexp))
					 ,(cadr ashexp)))))
	     ((and (pair? (caddr sexp)) (eq? 'IF (caaddr sexp)))
	      (let ((ifexp (caddr sexp)))
		(translate-exp
		 indent
		 `(IF ,(cadr ifexp)
		      ,@(map (lambda (se) `(logbit? ,(cadr sexp) ,se))
			     (cddr ifexp))))))
	     ((and (pair? (caddr sexp))
		   (assq (caaddr sexp) '((LOGIOR OR) (LOGAND AND))))
	      => (lambda (pr)
		   (translate-exp
		    indent
		    (cons (cadr pr)
			  (map (lambda (se) `(logbit? ,(cadr sexp) ,se))
			       (cdaddr sexp))))))
	     (else (report "strange exp to LOGBIT?" sexp))))
	   ((NOT)
	    (cond ((symbol? (cadr sexp))
		   (out CONTLINE "/" (cadr sexp)))
		  (else
		   (out CONTLINE "/" "(")
		   (translate-exp (+ 2 indent)(cadr sexp))
		   (out CONTLINE ")"))))
	   ((AND OR XOR)
	    (cond
	     ((memq (cdr (assq (car sexp)
			       '((OR . #t) (AND . #f) (XOR . gleep!))))
		    (cdr sexp))
	      (translate-exp
	       indent (cdr (assq (car sexp)
				 '((OR . #t) (AND . #f) (XOR . gleep!))))))
	     (else
	      (infix-translate-exp
	       (lookup (car sexp)
		       '((OR . " + ") (AND . " * ") (XOR . " :+: ")))
	       indent
	       (remove (cdr (assq (car sexp) '((OR . #f) (AND . #t) (XOR . #f))))
		       (cdr sexp))))))
	   ((IF)
	    (translate-exp indent `(or (and ,(cadr sexp) ,(caddr sexp))
				     (and ,(invertify (cadr sexp))
					  ,(cadddr sexp)))))
	   ((SYNCH:PRE)
	    (let ((expn (signal-table:get-next-fun *block* (cadr sexp))))
	      (if (not expn) (report "returned #f" sexp))
	      (translate-exp indent expn)))
	   ((CASE CASEV)
	    (let ((cv (cadr sexp)))
	      (translate-exp indent
			   (cons
			    'or (map
				 (lambda (clause)
				   `(and
				     ,(cond ((eq? 'else (car clause))
					     (cond
					      ((car (last-pair clause))
					       (report "else not #f?" clause)))
					     #t)
					    ((= 1 (length (car clause)))
					     `(= ,(case-eval (caar clause)) ,cv))
					    (else
					     `(or ,@(map (lambda (v)
							   `(= ,(case-eval v) ,cv))
							 (car clause)))))
				     ,@(if (and
					    (pair? (cadr clause))
					    (eq? 'and (caadr clause)))
					   (cdadr clause)
					   (list (cadr clause)))))
				 (cddr sexp))))))
	   ((CHECK=)
	    (translate-exp indent (cadddr sexp)))
	   ((SYNCH:CHAT NUMBER-CHECK)
	    (out CONTLINE "VCC"))
	   ;;((SYNCH:PRE) (translate-exp indent (macroexpand-1 sexp)))
	   (else (cond ((and (list? sexp) (= 2 (length sexp))
			     (apply signal-table:row-retrieve sexp))
			(translate-exp indent (cadr sexp)))
		       ((pair? (car sexp)) ;computed function
			(out indent "(*(")
			(translate-exp (+ 3 indent) (car sexp))
			(out CONTLINE "))")
			(out (+ 2 indent))
			(infix-translate-exp #\, (+ 2 indent) (cdr sexp)))
		       (else
			(out CONTLINE (car sexp))
			(infix-translate-exp #\, (+ 2 indent) (cdr sexp)))))))))

(define (optionally-parenthesize x indent)
    (cond ((and (pair? x) (not (memq (car x)
				     '(NOT BIT-EXTRACT BIT-FIELD LOGBIT))))
	   (out CONTLINE #\()
	   (translate-exp (+ 1 indent) x)
	   (out CONTLINE #\)))
	  (else (translate-exp indent x))))

(define (infix-translate-exp op indent exps)
  (cond ((eqv? #\, op)
	 (out CONTLINE #\()
	 (cond ((not (null? exps))
		(translate-exp indent (car exps))
		(set! exps (cdr exps))))
	 (for-each
	  (lambda (x)
	    (out CONTLINE op #\ )
	    (translate-exp indent x))
	  exps)
	 (out CONTLINE #\)))
	(else
	 (cond ((not (null? exps))
		(optionally-parenthesize (car exps) indent)
		(set! exps (cdr exps))))
	 (cond ((and (string? op)
		     (char=? #\  (string-ref op 0))
		     (some (lambda (x)
			     (and (pair? x)
				  (not (memq (car x) '(NOT LOGBIT?)))))
			   exps))
		(for-each
		 (lambda (x)
		   (out indent op)
		   (optionally-parenthesize
		    x (+ (if (char? op) 1 (string-length op)) indent)))
		 exps))
	       (else
		(for-each
		 (lambda (x)
		   (out CONTLINE op)
		   (optionally-parenthesize
		    x (+ (if (char? op) 1 (string-length op)) indent)))
		 exps))))))

(define (translate-= indent num1 arg2)
  (set! num1 (ceval num1))
  (set! arg2 (ceval arg2))
  (cond ((and (number? arg2) (not (number? num1)))
	 (let ((tmp num1))
	   (set! num1 arg2)
	   (set! arg2 tmp))))
  (let ((vs? (and (symbol? arg2) (vector-bounds arg2))))
    (cond ((not (number? num1))
	   (report "comparing non number?" num1))
	  ((and (not vs?)
		(not (and (list? arg2) (pair? arg2)
			  (memq (car arg2) '(BIT-EXTRACT BIT-FIELD)))))
	   (report "comparing non-vectorized signal?" (list num1 arg2)))
	  (else
	   (fluid-let ((*index* #f))
	     (translate-exp
	      indent (let* ((bounds (or vs?
					(list (car (vector-bounds (cadr arg2)))
					      (caddr arg2)
					      (+ -1 (cadddr arg2)))))
			    (first (cadr bounds))
			    (last (caddr bounds))
			    (ander '())
			    (incer (sign (- last first)))
			    (v (if vs? arg2 (cadr arg2))))
		       (do ((i first (+ incer i)))
			   ((positive? (* incer (- i last))))
			 ;;(print '= 'ind i '=> (- i (min first last)))
			 (set! ander (cons (if (logbit? (- i (min first last))
							num1)
					       `(logbit? ,i ,v)
					       `(not (logbit? ,i ,v)))
					   ander)))
		       (cons 'and
			     (if (positive? incer) ander
				 (reverse ander))))))))))

;;(print 'doing-if) (trace translate-assignment translate-top) (set! *qp-width* 333)
;;(trace-all "scm2mach.scm") (set! *qp-width* 333) (untrace out out-pins translate-name lookup registered-input? numbers-in-expression? ifable? infix-translate-exp) ; translat!e-exp
