;;;; scm2vrlg.scm: Program for translating SIMSYNCH code to VERILOG -*-scheme-*-
;;; Copyright (C) 1991, 1992, 1993, 1995, 1996, 1997, 1998 Aubrey Jaffer.

;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;; $Header: /usr/local/cvsroot/synch/scm2vrlg.scm,v 1.76 1999/01/13 05:06:08 jaffer Exp $

;;(require 'common-list-functions)
(require 'string-search)
(require 'fluid-let)
(require 'printf)
;;(require 'debug) (set! *qp-width* 100) (define qreport qpn)

;;; REPORT an error or warning
(define report
  (lambda args
    (display *verilog-output-name*)
    (display ": In function `")
    (display *current-signal-name*)
    (display "': ")
    (display *verilog-output-line-number*)
    (display ": warning: ")
    (apply qreport args)))
(define qreport
  (lambda args
    (for-each (lambda (x) (write x) (display #\ )) args)
    (newline)))

;;; This allows us to test without generating files
(define *verilog-output* (current-output-port))
(define *verilog-output-name* "?")
(define *acf-output* (current-output-port))

;;; *index* is the bit of an integer -- not a subscript.
(define *index* #f)			;dynamic variable for looping vects.
(define *range* #f)			;dynamic variable for numeric constants.
(define *current-signal-name* #f)
(define *verilog-output-line-number* 0)
(define CONTLINE -80)
(define NONE "")

(define (invertify sexp)
  (cond ((and (pair? sexp) (eq? 'NOT (car sexp)))
	 (cadr sexp))
	((boolean? sexp) (not sexp))
	(else (list 'NOT sexp))))
(define (negate pol sexp)
  (if pol (invertify sexp) sexp))

;;; OUT indents and displays the arguments
(define (out indent . args)
  (cond ((>= indent 0)
	 (newline *verilog-output*)
	 (set! *verilog-output-line-number* (+ 1 *verilog-output-line-number*))
	 (do ((j indent (- j 8)))
	     ((> 8 j)
	      (do ((i j (- i 1)))
		  ((>= 0 i))
		(display #\  *verilog-output*)))
	   (display #\	 *verilog-output*))))
  (for-each (lambda (a)
	      (cond ((eq? '? a) (display "?" *verilog-output*))
		    ((symbol? a)
		     (translate-name a *verilog-output*))
		    (else (display a *verilog-output*))))
	    args))

;;; TRANSLATE-NAME removes or translates characters from name and prints to port
(define (translate-name name port)
  (define visible? #f)
  (define bounds-spec? #f)
  (for-each
   (lambda (c)
     (let ((tc (cond ((char-alphabetic? c) c)
		     ((char-numeric? c) c)
		     ((char=? c #\.)
		      (cond (bounds-spec? (set! bounds-spec? #f) #\:)
			    (else #f)))
		     ((char=? c #\[) (set! bounds-spec? #t) #\[)
		     ((char=? c #\]) (set! bounds-spec? #f) #\])
		     ((char=? c #\-) #\_)
		     ((char=? c #\_) #\_)
		     ((char=? c #\?) "_P")
		     ((char=? c #\:) (cond (bounds-spec? #\:)
					   (else #\_)))
		     (else #f))))
       (cond (tc (set! visible? #t)
		 (display tc port)))))
   (string->list (symbol->string name)))
  (if (not visible?) (report "C-invisible symbol?" name)))

(define (outcomment indent str)
  (out indent "//" str)
  (out indent))

(define (vrlg:revectorize sym ind1 ind2)
  (revectorize ":" sym ind1 ind2))

(define (descmfilify file)
  (let ((sl (string-length file)))
  (cond ((< sl 4) file)
	((string-ci=? (substring file (- sl 4) sl) ".scm")
	 (substring file 0 (- sl 4)))
	(else file))))

;;; TRANSLATE files.
(define (translate-design-tag-from-files design tag files)
  (for-each (lambda (f) (translate1 f design tag ".v")) files))

;;; TRANSLATE1 - translate file.scm to file.suffix
(define (translate1 file design tag suffix)
  (define ofile
    (string-append (symbol->string design) "-" (symbol->string tag) suffix))
  (printf "translating %s -> %s for ptag %s\n" file ofile tag)
  (set! *verilog-output* (open-file ofile "w"))
  (set! *verilog-output-name* ofile)
  (set! *verilog-output-line-number* 0)
  (translate-tag design tag)
  (close-output-port *verilog-output*)
  (set! *verilog-output* (current-output-port)))

(define (numbers-in-expression? tree)
  (let walk ((tree tree))
    (cond ((number? tree))
	  ((not (pair? tree)) #f)
	  ((not (list? tree)) #f)
	  ((memq (car tree) '(logbit? =)) #f) ; ash bit-extract
	  ((some number? (cdr tree)))
	  (else (some walk (cdr tree))))))

(define (registered-input? name)
  (and (symbol? name)
       (let ((name (symbol->string name)))
	 (and (> (string-length name) 4)
	      (string=? "pre:" (substring name 0 4))
	      (string->symbol (substring name 4 (string-length name)))))))

(define (generate-verilog-header design tag)
  (require 'posix-time)
  (let* ((ptag (or (ptag->pld tag) "PALCE630"))
	 (signature (ptag->sig tag)))
    (newline *verilog-output*)
    (fprintf *verilog-output* "//TITLE     %s\n" (symbol-append design '- tag))
    (fprintf *verilog-output* "//PATTERN   %s\n" (or time-stamp:configuration ""))
    (fprintf *verilog-output* "//REVISION  %a\n" time-stamp:revision)
    (and time-stamp:author
	 (fprintf *verilog-output* "//AUTHOR    %s\n" time-stamp:author))
    (and time-stamp:company
	 (fprintf *verilog-output* "//COMPANY   %s\n" time-stamp:company))
    (fprintf *verilog-output* "//DATE      %s\n" (ctime time-stamp:time))
    (fprintf *verilog-output* "//CHIP  _%s  %s\n" tag ptag)
    (cond ((and signature
		(> (string-length ptag) 4)
		(string-ci=? (substring ptag 0 4) "mach"))
	   (if (not (eqv? signature (logand #xffffffff signature)))
	       (report "signature value out of range"
		       (number->string signature 16))
	       (fprintf *verilog-output* "//SIGNATURE #h%x\n" signature))))))

(define (out-decl type args)
  (if (pair? args)
      (let loop ((args args)
		 (iarg 1)
		 (last-bnds #f))
	(if (null? args)
	    (out CONTLINE #\;)
	    (apply
	     (lambda (nm . bnds)
	       (set! bnds (butlast bnds 1))
	       (cond ((equal? bnds last-bnds)
		      (out CONTLINE #\,)
		      (if (zero? (modulo iarg 6))
			  (out 10 nm)
			  (out CONTLINE #\space nm))
		      (loop (cdr args) (+ 1 iarg) bnds))
		     (else
		      (if last-bnds (out CONTLINE #\;))
		      (out 3 type)
		      (if (and (pair? bnds) (car bnds))
			  (out CONTLINE " [" (car bnds)
			       #\: (cadr bnds) #\]))
		      (out CONTLINE " " nm)
		      (cond ((and (pair? bnds) (pair? (cddr bnds)))
			     (out CONTLINE " [")
			     (translate-exp 5 NONE (caddr bnds))
			     (out CONTLINE #\:)
			     (translate-exp 5 NONE (cadddr bnds))
			     (out CONTLINE #\])))
		      (loop (cdr args) 1 bnds))))
	     (car args))))))

(define (merge-bounds lst alst)
  (let ((pair (assq (car lst) alst)))
    (cond ((not pair) (cons lst alst))
	  ((null? (cdr pair))
	   (report 'multiple-declaration-of pair)
	   alst)
	  ((null? (cdr lst))
	   (report 'multiple-declaration-of lst)
	   alst)
	  ((not (cadr pair))
	   (report 'replicated 'array? lst))
	  ((< (cadr pair) (caddr pair))
	   (set-car! (list-tail pair 1) (min (cadr pair) (cadr lst)))
	   (set-car! (list-tail pair 2) (max (caddr pair) (caddr lst)))
	   (set-car! (list-tail pair 3) (append (cadddr pair) (cadddr lst)))
	   alst)
	  (else
	   (set-car! (list-tail pair 1) (max (cadr pair) (cadr lst)))
	   (set-car! (list-tail pair 2) (min (caddr pair) (caddr lst)))
	   (set-car! (list-tail pair 3) (append (cadddr lst) (cadddr pair)))
	   alst))))

(define (get-bounds name)
  (let ((vb (vector-bounds name)))
    (if vb (append vb (list (list name))) (list name))))

(define (call-with-classified-pins tag block proc)
  (let ((outputs '())
	(inouts '())
	(inputs '())
	(registers '()))
    ((pin-table 'for-each-row)
     (lambda (pin-row)
       (apply 
	(lambda (ptag name pin-type pin-number output-enable paired-input)
	  (let ((lst (get-bounds name)))
	    (cond
	     ((or (not block) (signal-table:row-retrieve block name))
	      (case pin-type
		((output)
		 (set! outputs (merge-bounds lst outputs))
		 (set! registers (merge-bounds lst registers)))
		((input) 
		 (set! inputs (merge-bounds lst inputs)))
		((input/output)
		 (cond ((and (eq? paired-input name)
			     (not (eq? #t output-enable)))
			(set! registers (merge-bounds lst registers))
			(set! paired-input (symbol-append paired-input '- 'pin))
			(set! inouts (merge-bounds
				      (get-bounds paired-input)
				      inouts)))
		       (else
			(set! registers (merge-bounds lst registers))
			(set! inouts (merge-bounds
				      (get-bounds paired-input)
				      inouts)))))
		((internal)
		 (let* ((shape (array-bounds name))
			(newb (if (pair? shape)
				  (if (null? (cdr lst))
				      (append lst '(#f #f) shape)
				      (append lst shape))
				  lst)))
		   (cond (block
			  (if (used-by-other-blocks? name block)
			      (set! outputs (merge-bounds newb outputs)))
			  (set! registers (merge-bounds newb registers))))))
		((wire-and wire-or)
		 (report "found wire" pin-row))))
	     ((used-in-table:get-block name block)
	      (set! inputs (merge-bounds lst inputs))))))
	pin-row))
     tag)
    (proc outputs inouts inputs registers)))

(define (translate-constant-definition var def)
  (set! *current-signal-name* var)
  (out 3 "parameter ")
  (translate-name var *verilog-output*) ;name
  (out CONTLINE " = (")
  (if (boolean? def)
      (out CONTLINE (if def #\1 #\0))
;;;   (translate-exp 8 "\\" def)	; Uncomment this for silos
      (translate-exp 8 #f def)		; Uncomment this for leonardo
      )
  (out CONTLINE ");"))

(define (generate-module-header design tag block)
  (call-with-classified-pins
   tag block
   (lambda (outputs inouts inputs registers)
     (define block-controls ((block-table 'row:retrieve) block))
     (and block-controls
	  (print 'module block 'clock= (car block-controls)
		 'reset= (cadr block-controls))
	  (pretty-print
	   `(dump:spew
	     ,@(map (lambda (sig)
		      (case (length sig)
			((1) (car sig))
			((4)
			 (set! sig (cadddr sig))
			 (case (length sig)
			   ((1) (car sig))
			   (else `(dump:cat-bytes ,@sig))))
			(else (report 'funny-sig sig))))
		    (append inputs inouts outputs)))))
     (instantiate 'module (or block design) inputs inouts outputs)
     (cond (block
	    (out 0)
	    (outcomment 0 " Constants")
	    ((constant-table 'for-each-row)
	     (lambda (constant-row)
	       (apply
		(lambda (name value)
		  (cond ((and (not (used-in-table:get-block name block))
			      (not (boolean? value))))
			((or (boolean? value) (number? value))
			 (translate-constant-definition name value))))
		constant-row)))))
     (out 0)
     (out-decl "input" inputs)
     (out-decl "output" outputs)
     (out 0)
     (outcomment 0 "Bidirectional signals")
     (out-decl "inout" inouts)
     (out-decl "tri" inouts)
     (cond (block
	    (out 0)
	    (outcomment 0 "Registers")
	    (out-decl "reg" registers))))))

(define (generate-module-footer design tag block)
  (out 0)
  (out 0 "endmodule // " block))

(define (generate-standard-includes design tag)
  (display "\n\n//  Verilog includes\n" *verilog-output*)
  (out 0 "`include " "\"logical.v\"")
  (out 0))

;;; TRANSLATE-TAG - translate top level forms.
(define (translate-tag design tag)
  (generate-verilog-header design tag)
  (generate-module-header design tag #f)
  ((block-table 'for-each-row)
   (lambda (block-row)
     (apply
      (lambda (block-name clock reset)
	(call-with-classified-pins
	 tag block-name
	 (lambda (outputs inouts inputs registers)
	   (instantiate block-name
			(symbol-append tag '- block-name)
			inputs inouts outputs))))
      block-row)))
  (generate-module-footer design tag #f)
  ((block-table 'for-each-row)
   (lambda (block-row)
     (apply
      (lambda (block-name clock reset)
	;;(out 4 "// Block " block-name)
	(out 0)
	(generate-module-header design tag block-name)
	(generate-standard-includes design tag)
	((signal-table 'for-each-row)
	 (lambda (signal-row)
	   (apply
	    (lambda (block name defining-ptag signal-type
			   initial-state next-function order)
	      (and (or (not defining-ptag) (eq? tag defining-ptag))
		   (case signal-type ((macro) (out 3 "wire " name ";")))))
	    signal-row))
	 block-name)
	((signal-table 'for-each-row)
	 (lambda (signal-row)
	   (apply
	    (lambda (block name defining-ptag signal-type
			   initial-state next-function order)
	      (set! initial-state (ceval initial-state))
	      (and
	       (or (not defining-ptag) (eq? tag defining-ptag))
	       (or (pin-table:row-retrieve tag name)
		   (eq? signal-type 'macro))
	       (case signal-type
		 ((macro)
		  (out 3 "assign " name " = ")
		  (translate-exp 8 NONE next-function)
		  (out CONTLINE ";"))
		 ((registered registered-input combinatorial)
		  (apply
		   (lambda (pin-type pin-number output-enable paired-input)
		     (cond ((and (eq? paired-input name)
				 (not (eq? #t output-enable)))
			    (set! paired-input
				  (symbol-append paired-input '- 'pin))))
		     (cond ((and paired-input
				 (not (eq? paired-input #t))
				 (not (eq? output-enable #t)))
			    (out 3 "assign " paired-input " = ")
			    (translate-exp 8 NONE output-enable)
			    (out CONTLINE " ? " name " : "
				 (let ((bnds (vector-bounds name)))
				   (if bnds (+ 1 (abs (apply - (cdr bnds)))) 1))
				 "'bz;")))
		     (cond ((not (eq? paired-input #t))
			    (fluid-let ((*block* block))
			      (cond
			       ((or (number? initial-state)
				    (boolean? initial-state))
				(out 3 "always @ (posedge " clock
				     " or " " negedge " reset ") begin")
				(translate-top
				 `(define ,name (if (not ,reset)
						    ,initial-state
						    ,next-function))))
			       ((equal? name initial-state)
				(out 3 "always @ (posedge " clock ") begin")
				(qreport "Input-register " name)
				(translate-top `(define ,name ,next-function)))
			       (else
				(report "funny initial value for" name ':
					initial-state)
				(out 3 "always @ (posedge " clock
				     " or " " negedge " reset ") begin")
				(translate-top
				 `(define ,name (if (not ,reset)
						    #t
						    ,next-function))))))))
		     (out 3 "end //always")
		     (out 3))
		   (cddr (pin-table:row-retrieve tag name)))))))
	    signal-row))
	 block-name)
	;;(out 4 " // End Block " block-name)
	(generate-module-footer design tag block-name)
	(out 0))
      block-row)))
  (out 0)
  (generate-stimulus design tag))

(define (instantiate design tag inputs inouts outputs)
  (define loop (lambda (i args)
		 (cond ((pair? args)
			(let ((a (car args)))
			  (out (if (zero? (modulo i 6)) 6 CONTLINE) (car a))
			  (if (pair? (cdr args))
			      (out CONTLINE ", "))
			  (loop (+ i 1) (cdr args)))))))
  (define outsigs (lambda (name sigs)
		    (cond ((not (null? sigs))
			   (out 12 "//module " name)
			   (loop 0 sigs)))))
  (out 0)
  (out (if (eqv? 'module design) 0 3) design " " tag "(")
  (outsigs "inputs" inputs)
  (if (pair? inouts) (out CONTLINE ","))
  (outsigs "inouts" inouts)
  (if (pair? outputs) (out CONTLINE ","))
  (outsigs "outputs" outputs)
  (out CONTLINE ");"))

(define (generate-stimulus design tag)
  (call-with-classified-pins
   tag #f
   (lambda (outputs inouts inputs registers)
     (out 0 "`ifdef exemplar")
     (out 0 "// No stimulus module")
     (out 0 "`else")
     (out 0 "module " 'stimulus ";")
     (out 0)
     (outcomment 0 "module inputs")
     (out-decl "reg" inputs)
     (out 0)
     (outcomment 0 "module outputs")
     (out-decl "wire" outputs)
     (out 0)
     (outcomment 0 "module bidirectional signals")
     (out-decl "wire" inouts)
     (instantiate design tag inputs inouts outputs)
     (out 0)
     (out 0 'initial)
     (out 0 'begin)
     (for-each
      (lambda (input)
	(translate-assignment (car input) 3 #f 0))
      inputs)
     (out 0 'end)
     (out 0 "endmodule // " 'stimulus)
     (out 0 "`endif")
     (out 0))))

(define (for-each-index proc bounds)
  (let* ((first (cadr bounds))
	 (last (caddr bounds))
	 (range (- last first)))
    (if (positive? range)
	(do ((i 0 (+ 1 i))
	     (j first (+ 1 j)))
	    ((> i range))
	  (fluid-let ((*index* i))
	    ;;(print '+range i j)
	    (proc j)))
	(do ((i (- range) (+ -1 i))
	     (j first (+ -1 j)))
	    ((negative? i))
	  (fluid-let ((*index* i))
	    ;;(print '-range i j)
	    (proc j))))))

(define (for-each-value proc bounds)
  (let* ((first (cadr bounds))
	 (last (caddr bounds))
	 (range (- last first)))
    (if (not (zero? (min first last)))
	(report "mapping over non-consecutive values" bounds))
    (set! range (ash 2 (abs range)))
    (do ((i 0 (+ 1 i)))
	((>= i range))
      (fluid-let ((*index* i))
	(proc i)))))

;;; TRANSLATE-TOP - translate top level form sexp.
(define (translate-top sexp)
  (cond ((symbol? sexp) (set! *current-signal-name* sexp))
	((and (pair? sexp) (eq? (car sexp) 'QUOTE))
	 (set! *current-signal-name* (cadr sexp)))
	((string? sexp) (outcomment 0 sexp))
	((not (pair? sexp))
	 (report "top level atom?" sexp))
	(else (case (car sexp)
		((define) (translate-assignment (cadr sexp) 3 #f (caddr sexp)))
		(else (report "statement not in procedure" sexp)))
	      (set! *current-signal-name* #f)))
  (out 0))

(define (ifable? rhs)
  (and (pair? rhs)
       (case (car rhs)
	 ((CASE CASEV IF AND NOT ARRAY-SET LOGASET) #t) ; NUMBER-OR OR
	 ((SYNCH:PRE) (ifable? (signal-table:get-next-fun *block* (cadr rhs))))
	 (else #f))))

(define (translate-assignment lhs indent pol rhs)
  (set! *current-signal-name* lhs)
  (cond ((ifable? rhs)
	 (case (car rhs)
	   ((NOT)
	    (translate-assignment lhs indent (not pol) (cadr rhs)))
	   ((SYNCH:PRE)
	    (let ((expn (signal-table:get-next-fun *block* (cadr rhs))))
	      (if (not expn) (report "returned #f" rhs))
	      (translate-exp indent NONE expn)))
	   ((CASE CASEV) (out indent "case (")
	    (translate-exp (+ 6 indent) NONE (cadr rhs))
	    (out CONTLINE ")")
	    (for-each (lambda (clause)
			(translate-case-clause lhs (+ 2 indent) pol clause))
		      (cddr rhs))
	    (out indent "endcase"))
	   ((IF) (out indent "if (")
	    (translate-exp (+ 2 indent) NONE (cadr rhs))
	    (out CONTLINE ")")
	    (translate-assignment lhs (+ 2 indent) pol (caddr rhs))
	    (cond ((not (null? (cdddr rhs)))
		   (out indent "else")
		   (translate-assignment lhs (+ 2 indent) pol (cadddr rhs)))
		  (else (report 'dangling-if? lhs rhs))))
	   ((AND) (translate-assignment
		   lhs
		   indent
		   pol
		   (case (length rhs)
		     ((1) #t)
		     ((2) (cadr rhs))
		     (else `(if ,(butlast rhs 1) ,(car (last rhs 1)) #f)))))
	   ((array-set logaset)
	    (if (not (equal? lhs (cadr rhs)))
		(report 'array-set lhs rhs))
	    (let* ((bounds (vector-bounds lhs))
		   (name (if bounds (car bounds) lhs)))
	      (out indent name #\[)
	      (translate-exp (+ 10 indent) NONE (cadddr rhs))
	      (out CONTLINE #\])
	      (out (+ 4 indent) " <= ")	;name
	      (translate-exp (+ 8 indent) NONE (negate pol (caddr rhs)))
	      (out CONTLINE #\;)))
	   (else (report "ifable expression not translated" rhs))))
	;;((equal? lhs rhs) (out CONTLINE " ;"))
	(else
	 (out indent lhs " <= ")	;name
	 (translate-exp (+ 8 indent) NONE (negate pol rhs))
	 (out CONTLINE #\;))))

(define (case-eval dat clause)
  (cond ((and (pair? dat) (memq (car dat) '(UNQUOTE-SPLICING UNQUOTE)))
	 (if (not (constant-table:row-retrieve (cadr dat)))
	     (report (cadr dat) "case datum not constant in " clause))
	 (cadr dat))
	(else dat)))

(define (translate-case-clause lhs indent pol clause)
  (cond ((eq? 'else (car clause))
	 (let ((lc (car (last-pair clause))))
	   (cond ((and (pair? lc)
		       (memq (car lc) '(synch:chat synch:warn synch:error))))
		 (else
		  (if (not (member lc '(0 #f)))
		      (report "case else not (0) #f?" lc))
		  (out indent "default :")
		  (translate-assignment lhs (+ 2 indent) pol lc)))))
	((= 1 (length (car clause)))
	 (let ((dat (caar clause)))
	   (out indent (case-eval dat clause) " :"))
	 (translate-assignment lhs (+ 2 indent)
			       pol (car (last-pair clause))))
	(else
	 (for-each (lambda (match)
		     (translate-case-clause
		      lhs indent pol (cons (list match) (cdr clause))))
		   (car clause)))))

;;; LOOKUP - translate from table or return arg as string
(define (lookup arg tab)
  (let* ((p (assq arg tab))
	 (l (if p (cdr p) arg)))
    (if (symbol? l) (symbol->string l) l)))

;;; TRANSLATE-EXP - translate expression
;;; SEP is a string to output before each newline, or #f to inhibit
;;; any newlines.
(define (translate-exp indent sep sexp)
  (cond ((not (pair? sexp))		;atoms
	 (cond ((string? sexp)
		(out CONTLINE #\" sexp #\"))
	       ((boolean? sexp)
		(out CONTLINE (case sexp ((#f) "0") ((#t) "1"))))
	       ((and *index* (symbol? sexp) (vector-bounds sexp))
		(translate-exp indent sep (if (negative? *index*)
					      #f
					      `(logbit? ,*index* ,sexp))))
	       ((and *index* (integer? sexp))
		(out CONTLINE (if (and (not (negative? *index*))
				       (logbit? *index* sexp))
				  "1" "0")))
	       ((symbol? sexp) (out CONTLINE sexp))
	       ((number? sexp)
		(if *range*
		    (let* ((num (if (positive? *range*)
				    (bit-reverse *range* sexp)
				    sexp))
			   (str (number->string num 2)))
		      (out CONTLINE
			   (number->string (abs *range*))
			   "'b"
			   (make-string (- (abs *range*) (string-length str))
					#\0)
			   str))
		    (out CONTLINE (number->string sexp))))
	       (else (report "unexpected datatype:" sexp)
		     (out CONTLINE sexp))))
	(else
	 (case (car sexp)
	   ;; These vectorized functions are for multiple bit signals:
	   ((=)
	    (cond ((not (= 2 (length (cdr sexp))))
		   (report "= should take two arguments" sexp)))
	    (translate-= indent sep (cadr sexp) (caddr sexp)))
	   ((ZERO?)
	    (translate-= indent sep 0 (cadr sexp)))
	   ((LOGNOT)
	    (cond ((symbol? (cadr sexp))
		   (out CONTLINE "~")
		   (translate-exp (+ 2 indent) sep (cadr sexp)))
		  (else
		   (out CONTLINE "~" "(")
		   (translate-exp (+ 2 indent) sep (cadr sexp))
		   (out CONTLINE ")"))))
	   ((LOGIOR LOGAND LOGXOR + - * / modulo quotient)
	    (infix-translate-exp
	     (lookup (car sexp)
		     '((LOGIOR . " | ") (LOGAND . " & ") (LOGXOR . " ^ ")
		       (+ . " + ") (- . " - ") (* . " * ") (/ . " / ")
		       (modulo . " % ") (quotient . " / ")))
	     indent
	     sep
	     (cdr sexp)))
	   ((NUMBER-OR) (infix-translate-exp " | " indent sep (cddr sexp)))
	   ((ASH)
	    (cond (*index* (fluid-let ((*index* (- *index* (caddr sexp))))
			     (translate-exp indent sep (cadr sexp))))
		  ((not (number? (caddr sexp)))
		   (infix-translate-exp " << " indent sep (cdr sexp)))
		  ((negative? (caddr sexp))
		   (infix-translate-exp " >> " indent sep
					(list (cadr sexp) (- (caddr sexp)))))
		  (#t
		   (infix-translate-exp " << " indent sep (cdr sexp)))
		  (else (report "ash not handled" sexp))))
	   ((BIT-EXTRACT BIT-FIELD)
	    (let* ((bounds (vector-bounds (cadr sexp)))
		   (lowbnd (if bounds (min (cadr bounds) (caddr bounds)) 0)))
	      (cond ((not bounds)
		     (report "bit extract of non-symbol?" sexp)))
	      (translate-exp
	       CONTLINE
	       sep
	       (vrlg:revectorize (cadr sexp)
				 (+ lowbnd (caddr sexp))
				 (+ lowbnd (cadddr sexp) -1)))))
	   ((LOGBIT?)
	    (cond
	     ((and (number? (cadr sexp)) (negative? (cadr sexp))
		   (translate-exp indent sep #f)))
	     ((symbol? (caddr sexp))
	      (let* ((bounds (vector-bounds (caddr sexp)))
		     (lowbnd (if bounds (min (cadr bounds) (caddr bounds)) 0))
		     (dvs (devectorize (caddr sexp) (+ lowbnd (cadr sexp)))))
		;;(if (not (zero? lowbnd)) (print 'logbit? sexp bounds lowbnd))
		(translate-exp CONTLINE sep dvs)
		(cond (dvs (out CONTLINE #\[)
			   ;;(translate-exp CONTLINE sep (+ lowbnd (cadr sexp)))
			   (out CONTLINE (+ lowbnd (cadr sexp)))
			   (out CONTLINE #\])))))
	     (else
	      (out CONTLINE 'LOGBIT?)
	      (infix-translate-exp #\, (+ 2 indent) sep (cdr sexp)))))
	   ((ARRAY-REF LOGAREF)
	    (translate-logaref indent sep (cadr sexp) (cddr sexp)))
	   ((NOT)
	    (out CONTLINE "!")
	    (optionally-parenthesize (cadr sexp) (+ 1 indent) sep))
	   ((OR)
	    (cond ((memq #t (cdr sexp)) (translate-exp indent sep #t))
		  (else (infix-translate-exp
			 " || " indent sep (remove #f (cdr sexp))))))
	   ((XOR)
	    (infix-translate-exp " ^ " indent sep (remove #f (cdr sexp))))
	   ((&&)
	    (infix-translate-exp " && " indent sep (cdr sexp)))
	   ((AND)
	    (cond ((memq #f (cdr sexp)) (translate-exp indent sep #f))
		  (else (translate-and indent sep (remove #t (cdr sexp))))))
	   ((IF)
	    (optionally-parenthesize (cadr sexp) (+ 1 indent) sep)
	    (out CONTLINE " ? ")
	    (translate-exp (+ 4 indent) sep (caddr sexp))
	    (out CONTLINE " : ")
	    (translate-exp (+ 7 indent) sep (cadddr sexp)))
	   ((SYNCH:PRE)
	    (let ((expn (signal-table:get-next-fun *block* (cadr sexp))))
	      (if (not expn) (report "returned #f" sexp))
	      (translate-exp indent sep expn)))
	   ((CASE CASEV)
	    (let ((cv (cadr sexp)))
	      (translate-exp
	       indent sep
	       (cons
		'or (map
		     (lambda (clause)
		       `(and
			 ,(cond ((eq? 'else (car clause))
				 (cond
				  ((car (last-pair clause))
				   (report "else not #f?" clause)))
				 #t)
				((= 1 (length (car clause)))
				 `(= ,(case-eval (caar clause) clause) ,cv))
				(else
				 `(or ,@(map (lambda (v)
					       `(= ,(case-eval v clause) ,cv))
					     (car clause)))))
			 ,@(if (and
				(pair? (cadr clause))
				(eq? 'and (caadr clause)))
			       (cdadr clause)
			       (list (cadr clause)))))
		     (cddr sexp))))))
	   ((CHECK=)
	    (translate-exp indent sep (cadddr sexp)))
	   ((SYNCH:CHAT NUMBER-CHECK)
	    (out CONTLINE "1"))
	   ;;((SYNCH:PRE) (translate-exp indent sep (macroexpand-1 sexp)))
	   (else (cond ((and (list? sexp) (= 2 (length sexp))
			     (apply signal-table:row-retrieve sexp))
			(translate-exp indent sep (cadr sexp)))
		       ((pair? (car sexp)) ;computed function
			(out indent "(*(")
			(translate-exp (+ 3 indent) sep (car sexp))
			(out CONTLINE "))")
			(out (+ 2 indent))
			(infix-translate-exp #\, (+ 2 indent) sep (cdr sexp)))
		       (else
			(out CONTLINE (car sexp))
			(infix-translate-exp
			 #\, (+ 2 indent) sep (cdr sexp)))))))))

(define (translate-logaref indent sep array idxs)
  (let* ((bounds (vector-bounds array))
	 (name (if bounds (car bounds) array)))
    (out CONTLINE name #\[)
    (translate-exp indent sep (car idxs))
    (out CONTLINE #\])))

(define (optionally-parenthesize x indent sep)
    (cond ((and (pair? x) (not (memq (car x)
				     '(NOT BIT-EXTRACT BIT-FIELD LOGBIT))))
	   (out CONTLINE #\()
	   (translate-exp (+ 1 indent) sep x)
	   (out CONTLINE #\)))
	  (else (translate-exp indent sep x))))

(define (translate-and indent sep exps)
  (case (length exps)
    ((0) (translate-exp indent sep #t))
    ((1) (translate-exp indent sep (car exps)))
    (else
     (translate-exp
      indent sep `(if (&& ,@(butlast exps 1)) ,(car (last-pair exps)) #f)))))

(define (infix-translate-exp op indent sep exps)
  (cond ((eqv? #\, op)
	 (out CONTLINE #\()
	 (cond ((not (null? exps))
		(translate-exp indent sep (car exps))
		(set! exps (cdr exps))))
	 (for-each
	  (lambda (x)
	    (out CONTLINE op #\ )
	    (translate-exp indent sep x))
	  exps)
	 (out CONTLINE #\)))
	(else
	 (cond ((not (null? exps))
		(optionally-parenthesize (car exps) indent sep)
		(set! exps (cdr exps))))
	 (cond ((and (string? op)
		     (char=? #\  (string-ref op 0))
		     (some (lambda (x)
			     (and
			      (pair? x)
			      (not (memq
				    (car x)
				    '(NOT BIT-EXTRACT BIT-FIELD LOGBIT?)))))
			   exps))
		(for-each
		 (lambda (x)
		   (cond ((string? sep)
			  (out CONTLINE sep)
			  (out indent op))
			 (else (out CONTLINE op)))
		   (optionally-parenthesize
		    x (+ (if (char? op) 1 (string-length op)) indent) sep))
		 exps))
	       (else
		(for-each
		 (lambda (x)
		   (out CONTLINE op)
		   (optionally-parenthesize
		    x (+ (if (char? op) 1 (string-length op)) indent) sep))
		 exps))))))

(define (translate-= indent sep num1 arg2)
  (cond ((and (number? arg2) (not (number? num1)))
	 (let ((tmp num1))
	   (set! num1 arg2)
	   (set! arg2 tmp))))
  (infix-translate-exp " == " indent sep (list num1 arg2)))

;;(print 'doing-if) (trace translate-assignment translate-top) (set! *qp-width* 333)
;;(trace-all "/home/jaffer/synch/scm2vrlg.scm") (set! *qp-width* 333) (untrace out translate-name lookup registered-input? numbers-in-expression? ifable? infix-translate-exp merge-bounds get-bounds) ; translat!e-exp
